"""
Call gitlab-emulator's code to execute a job
"""
import sys
import glob
import os
import shutil
import tempfile
import subprocess
import time
import stat
import traceback

import requests
import zipfile

import yaml

from .common import gethostname, iswindows, generate_config
from .trace import TraceProxy
from .consts import NAME, VERSION
from unidecode import unidecode


def get_variable(job, *names, default=None):
    """
    Get a named variable from a job
    :param job:
    :param name:
    :param default:
    :return:
    """
    for var in job["variables"]:
        for name in names:
            if var["key"] == name:
                return var["value"]
    return default


def archive(trace, config, jobname, tempdir, build_folder, success):
    """
    Create a zip archive of the build artifacts if required
    :param trace: trace proxy
    :param config: global gitlab config data
    :param jobname: the job name
    :param tempdir: the temporary data folder
    :param build_folder: the root folder of this build
    :param success: True for successful builds
    :return:
    """
    # handle artifact upload
    job_config = config[jobname]
    if "artifacts" in job_config:
        when = "on_success"  # default
        if "when" in job_config["artifacts"]:
            when = job_config["artifacts"]["when"]

        if when != "always":
            if when == "on_success" and not success:
                return None
            if when == "on_failure" and success:
                return None

        if "paths" not in job_config["artifacts"]:
            return None

        excludes = []
        excludepatts = job_config["artifacts"].get("exclude", [])
        for item in excludepatts:
            if os.sep != '/':
                item = item.replace("/", os.sep)
            excludepatt = os.path.join(build_folder, item)
            excludes.extend(glob.glob(excludepatt, recursive=True))

        with trace.section("artifacts", "Prepare Artifacts"):
            trace.writeline("\nFinding artifacts...")
            if excludes:
                trace.writeline("\nExclude {} artifacts".format(len(excludes)))

            paths = job_config["artifacts"]["paths"]
            zipname = "archive.zip"
            if "name" in job_config["artifacts"]:
                # TODO expand variables for this
                zipname = job_config["artifacts"]["name"] + ".zip"

            zippath = os.path.join(tempdir, zipname)
            with zipfile.ZipFile(zippath, "w") as zipobj:
                for item in paths:
                    trace.writeline(".. finding {}".format(item))
                    # patterns are posix paths and globs, convert them to local on windows
                    if os.sep != '/':
                        item = item.replace("/", os.sep)
                    localpatt = os.path.join(build_folder, item)
                    matches = glob.glob(localpatt, recursive=True)

                    for include in matches:
                        if include in excludes:
                            continue
                        relpath = os.path.relpath(include, build_folder)
                        if os.path.isfile(include):
                            trace.writeline(".. match {}".format(include))
                            zipobj.write(include, relpath)
                        if os.path.isdir(include):
                            # recurse adding the whole folder
                            for root, _, files in os.walk(include):
                                for file in files:
                                    fname = os.path.join(root, file)
                                    relname = os.path.relpath(fname, build_folder)
                                    trace.writeline(".. match {}".format(relname))
                                    zipobj.write(fname, relname)
        return zippath
    return None


def clean(trace, folder, docker, attempts=10):
    """
    Totally delete the given folder regardless of permissions inside
    :param trace:
    :param folder:
    :param docker:
    :param attempts: retry this many times
    :return:
    """
    from gitlabemu import logmsg

    logmsg.info("Cleaning build {}".format(folder))
    # make everything deletable on windows
    if iswindows():
        for root, dirs, files in os.walk(folder):
            for item in files + dirs:
                path = os.path.join(root, item)
                if not os.access(path, os.W_OK):
                    # shutil.rmtree() will barf on ro files
                    os.chmod(path, stat.S_IWUSR)
                    # so lets delete it now
                    os.unlink(path)
    try:
        try:
            shutil.rmtree(folder)
        except:
            # failed, if we have docker, try to use that
            if docker:
                from gitlabemu import logmsg
                if iswindows():
                    logmsg.fatal("unable to clean {}".format(folder))
                else:
                    # try to use a busybox
                    parent = os.path.abspath(os.path.dirname(folder))
                    foldername = os.path.basename(folder)
                    subprocess.check_call(
                        ["docker", "run", "--rm",
                         "-w", parent,
                         "-v", parent + ":" + parent,
                         "busybox:latest",
                         "rm", "-rf", foldername
                         ])
                    return
            raise
    except Exception as err:
        # possibly this is windows virus scanning?
        if attempts > 0:
            trace.writeline("clean failed.. sleeping ".format(err, attempts))
            time.sleep(3 * (4 - attempts))
            clean(trace, folder, docker, attempts - 1)
        else:
            trace.writeline("Error cleaning after build {}".format(err))
            raise
    finally:
        trace.flush()


def trace_checkoutput(trace, cmdline, cwd):
    """
    Run a program and emit the trace
    :param cmdline:
    :param cwd:
    :return:
    """
    try:
        trace.write(subprocess.check_output(cmdline, cwd=cwd, stderr=subprocess.STDOUT))
    except subprocess.CalledProcessError as cpe:
        trace.writeline("Error running {}".format(cmdline))
        trace.write(cpe.output)
        raise


def run(runner, job, docker):
    """
    Execute the given job here using gitlab-emulator
    :param runner: the runner object
    :param job: the job response from the server
    :param docker: if True, we can use docker for housekeeping
    :return:
    """
    from gitlabemu import logmsg, configloader, errors

    trace = TraceProxy(runner, job)

    logmsg.FATAL_EXIT = False

    trace.writeline("Running on {} {} {}".format(gethostname(), NAME, VERSION))
    if docker:
        trace.writeline("Using Docker executor")
    else:
        trace.writeline("Using shell executor")

    tempdir = tempfile.mkdtemp(dir=runner.builds)
    build_dir = os.path.join(tempdir, get_variable(job, "CI_PROJECT_PATH"))
    build_dir_abs = os.path.abspath(build_dir)
    try:
        with trace.section("git", "Source Control"):
            os.makedirs(build_dir)
            # clone the git repo defined in the job
            git = job["git_info"]

            trace.writeline("Cloning project..")
            clonecmd = ["git", "clone", git["repo_url"], build_dir]
            depth = git.get("depth", None)
            if depth:
                reftype = git.get("ref_type", None)
                if reftype == "branch":
                    refspecs = git.get("ref_specs", [])
                    if refspecs:
                        trace.writeline(f"Cloning to depth: {depth}")
                        clonecmd.extend(["--depth", str(depth),
                                         "--branch", refspecs[0]])

            trace_checkoutput(trace, clonecmd, cwd=tempdir)

            # checkout the ref to build
            trace_checkoutput(trace, ["git", "checkout", "-f", get_variable(job, "CI_COMMIT_SHA", "CI_COMMIT_REF")], cwd=build_dir_abs)


        # config_file
        jobname = job["job_info"]["name"]
        config, derived_config = load_pipeline_config(trace, tempdir, build_dir_abs, job, jobname)

        if derived_config:
            trace.writeline("Dervied work folder is {}".format(build_dir_abs))

        # populate real vars
        for var in job["variables"]:
            # support 'file' variables
            if 'file' in var.keys() and var["file"]:
                filevar = tempfile.NamedTemporaryFile(dir=tempdir, delete=False )
                filevar.write(str.encode(var["value"]))
                name = var["key"]
                config["variables"][name] = filevar.name
                filevar.close()
            else:
                # fix 'None' values
                if var["value"] is None:
                    var["value"] = ""
                name = var["key"]
                config["variables"][name] = unidecode(var["value"])

        emulator_job = configloader.load_job(config, jobname)
        trace.emulator_job = emulator_job

        with trace.section("depend-artifacts", "Getting Artifacts"):
            runner.get_dependencies(trace, job, build_dir_abs)

        emulator_job.stdout = trace

        success = False
        error = False

        try:
            with trace.section("job", "Execute Job"):
                emulator_job.run()
            success = True
        except errors.GitlabEmulatorError:
            # the job failed
            try:
                emulator_job.abort()
            except OSError:
                pass
        except requests.HTTPError:
            error = True
            emulator_job.abort()

        if not error:  # success or failure but not some internal errors
            archive_file = archive(trace, config, jobname, tempdir, build_dir_abs, success)
            if archive_file:
                runner.upload(trace, job, archive_file)

        trace.writeline("Job complete success={}".format(success))
        return success
    except Exception as err:
        trace.writeline("Runner error: {}".format(err))
        error_type, error, tb = sys.exc_info()
        stacks = traceback.format_tb(tb)
        for x in stacks:
            trace.writeline(x)

    finally:
        clean(trace, tempdir, docker)


def load_pipeline_config(trace, tempdir, build_dir_abs, job, jobname):
    from gitlabemu import configloader
    loader = configloader.Loader()
    derived = False
    ci_cfg = get_variable(job, "CI_CONFIG_PATH")
    if ci_cfg:
        ci_file = os.path.join(build_dir_abs, ci_cfg)
    if os.path.exists(ci_file):
        loader.load(ci_file)

    config = loader.config

    if not ci_cfg or jobname not in loader.get_jobs():
        with trace.section("child-job", "Execute child job"):
            trace.writeline("Warning: CI_CONFIG_PATH is empty or unset")
            trace.writeline("Warning: gitlab-python-runner child-pipeline support is experimental")
            trace.writeline("Computing child job steps..")

            new_config = generate_config(job)
            ci_cfg = os.path.join(tempdir, "child-job.yml")
            with(open(ci_cfg, "w")) as new_config_fh:
                yaml.safe_dump(new_config, stream=new_config_fh)
            with(open(ci_cfg, "r")) as new_config_fh:
                for line in new_config_fh:
                    trace.writeline("#" + line)
        config = configloader.read(ci_cfg, topdir=build_dir_abs)
        config["variables"]["CI_PROJECT_DIR"] = build_dir_abs
        config[".gitlab-emulator-workspace"] = build_dir_abs

        derived = True

    return config, derived



"""
OO Wrapper around requests
"""
import sys
from typing import Union, Text, Optional, Any

import requests
from requests.models import Response

from . import consts
from .common import trace_http_request, trace_http_response, trace_http_response_raw


class Session(requests.Session):
    """
    A HTTP user agent complete with cookies
    """
    def __init__(self):
        super(Session, self).__init__()
        pyver = sys.version_info
        pyver_str = "; python {}.{}".format(pyver.major, pyver.minor)
        self.headers["User-Agent"] = consts.USER_AGENT + " " + pyver_str

    def post(self, url, **kwargs) -> Response:
        jdata = kwargs.get("json")
        trace_http_request("post", url, jdata)
        resp = super().post(url, **kwargs)
        trace_http_response_raw(resp)
        return resp

    def json(self, resp):
        data = resp.json()
        trace_http_response(data)
        return data



"""
Functions for adding gitlab-runner.py as a service.
"""
from __future__ import print_function
import os
import shutil
from .common import parse_config


class BaseService(object):
    """
    Base service implementations
    """

    def detect(self):
        """
        Check that this service system is supported
        :return:
        """
        return False

    def is_installed(self, config):
        """
        :param config: path to the configuration used.
        :return: True if already installed
        """
        return False

    def install(self, config, user):
        raise NotImplementedError()

    def uninstall(self):
        raise NotImplementedError()

    def configure(self, configfile, user):
        """ write the shell config file
        """
        confdir = "/etc/gitlab-python-runner/"
        config = parse_config(configfile)
        if not os.path.exists(confdir):
            os.makedirs(confdir)
        assert configfile and config
        shutil.copyfile(configfile, os.path.join(confdir, "gitlab-runner.yml"))
        if user is None:
            user = "gitlab-runner"
        logfile = os.path.join(config["dir"], "runner.log")

        shellfile = "/etc/gitlab-python-runner/service.conf"

        with open(shellfile, "w") as outfile:
            outfile.write("""
RUNNER_USER={}
LOGFILE={}
export RUNNER_USER
export LOGFILE            
            """.format(user, logfile))
        return user


class SystemV(BaseService):
    """
    Install/Uninstall service on using SysV
    """
    START = "S90gitlab-python-runner"
    STOP = "K90gitlab-python-runner"

    INIT_SCRIPT = os.path.join(os.path.dirname(__file__), "gitlab-runner-init.sh")
    ETC_INIT = "/etc/init.d/gitlab-python-runner"

    def install(self, config, user):
        user = self.configure(config, user)
        print("Installing config {} as user {}".format(config, user))
        if os.path.exists(self.ETC_INIT):
            os.unlink(self.ETC_INIT)
        shutil.copyfile(self.INIT_SCRIPT, self.ETC_INIT)
        os.chmod(self.ETC_INIT, 0o755)
        for i in [2]:
            os.symlink(self.ETC_INIT, os.path.join("/etc/rc{}.d".format(i), self.START))
        for i in [0]:
            os.symlink(self.ETC_INIT, os.path.join("/etc/rc{}.d".format(i), self.STOP))

    def is_installed(self, config):
        return os.path.exists(os.path.join("/etc", "rc2.d", self.START))

    def uninstall(self):
        for i in [2]:
            filename = os.path.join("/etc/rc{}.d".format(i), self.START)
            if os.path.exists(filename):
                os.unlink(filename)
        for i in [0]:
            filename = os.path.join("/etc/rc{}.d".format(i), self.STOP)
            if os.path.exists(filename):
                os.unlink(filename)
        if os.path.exists(self.ETC_INIT):
            os.unlink(self.ETC_INIT)

    def detect(self):
        required = ["/etc/rc2.d", "/etc/init.d", "/etc/rc0.d"]
        for item in required:
            if not os.path.exists(item):
                return False
        return True


def get_installer():
    for test in [SystemV()]:
        if test.detect():
            return test
    return None



"""
Various functions shared between the other modules
"""
import socket
import platform
import zipfile
import os
import yaml
import json
import time
from yaml.constructor import ConstructorError


_trace_http = False
_trace_file = os.path.join(os.getcwd(), "http.trace")


def set_trace_http_requests(enable):
    global _trace_http
    _trace_http = enable


def _append_http_trace(message):
    """
    Append the http trace to the log file
    :param message:
    :return:
    """
    with open(_trace_file, "a") as fileh:
        print(str(message) + "---\n", file=fileh)


def trace_http_request(method, url, json_data):
    """
    Log the given json data
    :param method:
    :param url:
    :param json_data:
    :return:
    """
    if _trace_http:
        _append_http_trace("> {} {} json: {}".format(method.upper(), url, json.dumps(json_data)))


def trace_http_response(json_data):
    """
    Log the given json response
    :param json_data:
    :return:
    """
    if _trace_http:
        _append_http_trace("< json: {}".format(json.dumps(json_data)))

def trace_http_response_raw(resp):
    """
    Log the given json response
    :param resp:
    :return:
    """
    if _trace_http:
        _append_http_trace("< raw: status={}".format(resp.status_code))
        if resp.status_code >= 300:
            _append_http_trace("< raw: {}".format(str(resp)))


def gethostname():
    try:
        return socket.gethostname()
    except:
        return "unknown-hostname"


def iswindows():
    return platform.system() == "Windows"


def parse_config(configfile):
    with open(configfile, "r") as infile:
        try:
            config = yaml.load(infile, Loader=yaml.SafeLoader)
        except ConstructorError:
            # this file probably has unicode still in it, use the full loader
            config = yaml.load(infile, Loader=yaml.FullLoader)

    assert config

    assert "server" in config
    assert "dir" in config
    assert "executor" in config
    assert "token" in config
    assert "shell" in config

    return config


def save_config(configfile, data):
    with open("gitlab-runner.yml", "w") as outfile:
        yaml.safe_dump(data, outfile, indent=2)


class ZipFileEx(zipfile.ZipFile):
    """
    A variant of ZipFile that restores file permissions where supported
    """
    # inspired mostly from https://stackoverflow.com/a/39296577/148415

    def _extract_member(self, member, targetpath, pwd):
        """
        Extract a member from the zip file and restore permission bits, also ensure that
        the file is readable and writable by the owner
        :param member:
        :param path:
        :param pwd:
        :return:
        """
        if not isinstance(member, zipfile.ZipInfo):
            member = self.getinfo(member)
        ret_val = super(ZipFileEx, self)._extract_member(member, targetpath, pwd)
        if not iswindows():
            attr = member.external_attr >> 16
            os.chmod(ret_val, attr | 0o600)
        return ret_val


def generate_config(polled):
    """
    Given the polled job response, generate a gitlab-ci yaml file content
    :param polled:
    :return:
    """
    global_config = dict()
    job_config = dict()
    stage = polled["job_info"]["stage"]
    global_config["stages"] = [stage]
    job_config["stage"] = stage
    image = polled["image"]
    if image:
        job_config["image"] = image
    services = polled["services"]
    if services:
        global_config["services"] = services

    if "artifacts" in polled and polled["artifacts"] is not None:
        job_config["artifacts"] = {}
        for item in polled["artifacts"]:
            for element in ["name", "when", "paths", "reports", "expire_in", "exclude"]:
                if element in item:
                    value = item.get(element)
                    if value is not None:
                        job_config["artifacts"][element] = value

    for step in polled["steps"]:
        job_config[step["name"]] = step["script"]

    global_config[polled["job_info"]["name"]] = job_config

    return global_config



from distutils.core import setup
from GitlabPyRunner import consts

setup(
    name="gitlab-python-runner",
    version=consts.VERSION,
    description="Pure python gitlab-runner",
    author="Ian Norton",
    author_email="inorton@gmail.com",
    url="https://gitlab.com/cunity/gitlab-python-runner",
    packages=["GitlabPyRunner"],
    include_package_data=True,
    scripts=["gitlab-runner.py"],
    install_requires=[
        "pyyaml>=5.1",
        "requests>=2.23.0",
        "gitlab-emulator>=0.8.2",
        "requests-toolbelt>=0.9.1", 
        "unidecode>=1.1.1"],
    platforms=["any"],
    license="License :: OSI Approved :: MIT License",
    long_description="Gitlab compatible runner without Go or SSH"
)
